/* io/dns.c */
/* Non-blocking nameserver interface routines */

#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include "externs.h"

typedef unsigned char byte;

struct resolve {
   struct resolve *next;
   struct resolve *nextquery;
   char *hostname;
   int expire;
   unsigned int ip;
   unsigned int id:16;
   unsigned int state:8;
} *resolves=NULL, *queries=NULL;

// Macros for getting int16/int32 network byte-order numbers from a structure
// in memory. Pointer is advanced.

#define get16(ptr) (((unsigned int)*(ptr) << 8) | (((ptr)+=2)[-1]))
#define get32(ptr) (((unsigned int)*(ptr) << 24) | ((ptr)[1] << 16) | \
                    ((ptr)[2] << 8) | (((ptr)+=4)[-1]))

#define STATE_FINISHED	0
#define STATE_FAILED	1

#define NUM_TRIES	3

long resrecvbuf[(PACKETSZ+7) >> 2];


/* open the nameserver for querying */
int dns_open() {
  int s, on=1;

  res_init();
  if(!_res.nscount) {
    log_error("Resolver: No nameservers defined.");
    return -1;
  }

  /* open socket for nameserver communication */
  if((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("dns_open");
    return -1;
  } if(setsockopt(s, SOL_SOCKET, SO_BROADCAST, (char *)&on, sizeof(on)) < 0) {
    perror("dns_open: setsockopt");
    return -1;
  }

  return s;
}

void send_dns_request(struct resolve *rp, int type) {
  HEADER *hp;
  char buf[32], packet[PACKETSZ], *s;
  int r, i;

  if(type == T_A) {  /* look up host address */
    s=rp->hostname;
    log_io(tprintf("Resolver: Sent reverse authentication request for \"%s\".",
           rp->hostname));
  } else if(type == T_PTR) {  /* look up host name */
    sprintf(buf, "%u.%u.%u.%u.in-addr.arpa", ((byte *)&rp->ip)[3],
        ((byte *)&rp->ip)[2], ((byte *)&rp->ip)[1], ((byte *)&rp->ip)[0]);
    s=buf;
    log_io(tprintf("Resolver: Sent domain lookup request for \"%s\".",
           ipaddr(rp->ip)));
  } else
    return;

  if((r=res_mkquery(QUERY,s,C_IN,type,NULL,0,NULL,packet,PACKETSZ)) < 0) {
    perror("res_mkquery");
    return;
  }

  /* set query id */
  hp=(HEADER *)packet;
  hp->id=rp->id;

  /* send query to all nameservers */
  for(i=0;i<_res.nscount;i++)
    sendto(resolv, packet, r, 0, (struct sockaddr *)&_res.nsaddr_list[i],
           sizeof(struct sockaddr));
}

void failrp(struct resolve *rp) {
  if(rp->state == STATE_FINISHED)
    return;
  rp->state=STATE_FAILED;
  rp->expire=now+120;

  log_io("Resolver: Lookup failed.");
}

void passrp(struct resolve *rp, int ttl) {
  if(ttl <= 0)
    ttl=86400;
  rp->state=STATE_FINISHED;
  rp->expire=now+ttl;

  log_io(tprintf("Resolver: Lookup successful: %s (ttl=%d)",rp->hostname,ttl));
}

void dns_receive() {
  HEADER *hp;
  struct sockaddr_in from;
  struct resolve *rp;
  int ttl, len, r, fromlen=sizeof(from);
  short rr, datatype, class, qdatatype, qclass;
  char rdatalength, *end, *c, *s=(char *)resrecvbuf;
  char querystring[32], namestring[MAXDNAME];

  /* receive message on socket */
  if((len=recvfrom(resolv, resrecvbuf, PACKETSZ, 0,
     (struct sockaddr *)&from, &fromlen)) < 0) {
    perror("dns_receive");
    return;
  }

  /* packet is smaller than standard header size, or has empty body. */
  if(len <= sizeof(HEADER))
    return;

  // tc=1: Nameserver packet truncated.
  // qr=0: Query packet received on nameserver communication socket.
  // opcode != 0: Not a standard query. (1 = inverse request, 2 = server status)

  hp=(HEADER *)s;
  if(hp->tc || !hp->qr || hp->opcode)
    return;

  /* search for id in cache (id is echoed by nameserver) */
  for(rp=queries;rp;rp=rp->next)
    if(rp->id == hp->id)
      break;

  /* check if address has already been resolved */
  if(!rp || rp->state < 2)
    return;

  /* all values returned are in network byte order, so we must convert */
  hp->qdcount=ntohs(hp->qdcount);
  hp->ancount=ntohs(hp->ancount);
  hp->nscount=ntohs(hp->nscount);
  hp->arcount=ntohs(hp->arcount);

  log_io(tprintf("Received nameserver reply. (qd:%u an:%u ns:%u ar:%u)",
         hp->qdcount, hp->ancount, hp->nscount, hp->arcount));

  /* Process response */

  if(hp->rcode == NXDOMAIN)  /* Host not found */
    failrp(rp);
  if(hp->rcode != NOERROR)   /* Server failure or refusal */
    return;

  end=s+len;
  c=s+HFIXEDSZ;

  if(!hp->ancount)	/* No answers returned. */
    return;
  if(hp->qdcount != 1)	/* Reply does not contain one query. */
    return;
  if(c > end)		/* Reply too short. */
    return;

  /* construct expected query reply */
  sprintf(querystring, "%u.%u.%u.%u.in-addr.arpa", ((byte *)&rp->ip)[3],
          ((byte *)&rp->ip)[2], ((byte *)&rp->ip)[1], ((byte *)&rp->ip)[0]);

  /* expand compressed domain name and compare with query */
  if((r=dn_expand(s, s+len, c, namestring, MAXDNAME)) < 0)
    return;
  if(strcasecmp(querystring, namestring))  /* names do not match */
    return;

  c+=r;
  if(c+4 > end)  /* Query resource record truncated */
    return;

  qdatatype=get16(c);	// Data type = Domain name pointer
  qclass=get16(c);	// Class = Arpa Internet

  if(qclass != C_IN || qdatatype != T_PTR)  /* server response does not match */
    return;

  /* extract all resource records */
  for(rr = hp->ancount + hp->nscount + hp->arcount;rr;rr--,c+=rdatalength) {
    if(c > end)
      return;
    if((r=dn_expand(s, s+len, c, namestring, MAXDNAME)) < 0)
      return;

    /* get record data */
    c+=r;
    if(c+10 > end)  /* Resource record truncated */
      return;
    datatype=get16(c);
    class=get16(c);
    ttl=get32(c);
    rdatalength=get16(c);

    /* check that the query comes from the correct resource */
    if(strcasecmp(querystring, namestring) || datatype != qdatatype)
      continue;

    if(class != qclass)  /* Answered class doesn't match queried class */
      return;
    if(!rdatalength)  /* Zero size data */
      return;
    if(rdatalength+c > end)  /* Data length exceeds packet size */
      return;

    if(datatype == T_A) {  /* Reverse authentication */
      if(rdatalength != 4)  /* Unsupported rdata format for "A" type. */
        return;
      if(memcmp(&rp->ip, (int *)c, sizeof(int)))
        failrp(rp);
      else {
        passrp(rp, ttl);
        return;
      }
    } else if(datatype == T_PTR) {  /* Hostname Lookup */
      if((dn_expand(s, s+len, c, namestring, MAXDNAME)) > 0) {
        SPTR(rp->hostname, namestring);
        passrp(rp, ttl);
        return;
      }
    }
  }
}

/* called once per second */
void dns_events() {
  struct resolve *rp, *next, *last=NULL;

  for(rp=queries;rp;rp=next) {
    next=rp->nextquery;

    /* remove any finished or failed queries from list */
    if(rp->state < 2) {
      if(last)
        last->next=next;
      else
        queries=next;
      continue;
    }

    /* query has not expired yet */
    if(now < rp->expire) {
      last=rp;
      continue;
    }

    if(rp->state < NUM_TRIES+1) {
      rp->state++;
      rp->expire=now+3;
      send_dns_request(rp, T_PTR);
      last=rp;
      continue;
    }

    /* exceeded maximum number of retries */

    log_io(tprintf("Resolver: Query for %s timed out.", ipaddr(rp->ip)));
    failrp(rp);

    /* remove event from query list */
    if(last)
      last->next=next;
    else
      queries=next;
    rp->nextquery=NULL;
  }
}

/* Non-blocking gethostbyaddr(addr). Returns 0 if lookup is incomplete. */
char *dns_lookup(int ip)
{
  static unsigned short lookup_id=0;
  struct resolve *rp;

  /* if no nameserver present, just return address in dot notation */
  if(resolv == -1)
    return ipaddr(ip);

  /* search for a previously cached entry */
  for(rp=resolves;rp;rp=rp->next)
    if(rp->ip == ip)
      break;

  /* add a new record if not found */
  if(!rp) {
    rp=malloc(sizeof(struct resolve));
    rp->next=resolves;
    rp->state=0;
    rp->expire=0;
    rp->ip=ip;
    resolves=rp;
  }

  /* if the entry is new or has expired, send dns request */
  if(rp->state < 2 && rp->expire <= now) {
    rp->state=2;
    rp->expire=now+3;
    rp->id=++lookup_id;

    /* add entry to the list of queries */
    rp->nextquery=queries;
    queries=rp;

    send_dns_request(rp, T_PTR);
    return NULL;
  }

  if(rp->state == STATE_FINISHED)
    return rp->hostname;
  if(rp->state == STATE_FAILED)
    return ipaddr(ip);
  return NULL;
}
