/* comm/move.c */

#include "externs.h"

/* moves through an exit in room */
void do_move(dbref player, char *direction)
{
  dbref exit;
  int a, lock=0;

  /* check object type */
  if(Typeof(player) == TYPE_ROOM || Typeof(player) == TYPE_EXIT ||
     Typeof(player) == TYPE_ZONE) {
    notify(player, tprintf("Objects of type %s are not allowed to move.",
           typenames[Typeof(player)]));
    return;
  }

  if(RESTRICT_TELEPORT && !power(player,POW_TELEPORT) && hold_player(player)) {
    notify(player, "You may not move while carrying a player.");
    return;
  }

  /* First match all visible exits, then invisible ones */
  exit=match_thing(player, direction, MATCH_EXIT|MATCH_HOME|MATCH_QUIET);
  if(exit == NOTHING && (exit=match_thing(player, direction,
     MATCH_EXIT|MATCH_INVIS|MATCH_QUIET)) == NOTHING) {
    notify(player, "You can't go that way.");
    return;
  }

  /* homeward bound */
  if(exit == HOME) {
    if(db[player].location == db[player].link) {
      notify(player, "You are already home.");
      return;
    } if(RESTRICT_TELEPORT && !Builder(player) && !power(player,POW_TELEPORT)) {
      notify(player, "Yes, home.. what a special place to go back to.");
      return;
    }

    /* display messages */
    notify(player, "\e[1mThere's no place like home...\e[m");
    notify_except(db[player].location, player, tprintf("%s goes home.",
                  db[player].name));
    enter_room(player, HOME);
    return;
  }

  /* check current location's leave-lock */
  if(!could_doit(player, db[player].location, A_LLOCK)) {
    ansi_did_it(player, db[player].location, A_LFAIL, "You can't go that way.",
                A_OLFAIL, NULL, A_ALFAIL);
    return;
  }
  
  /* check lock on exit */
  if(db[exit].link == NOTHING || !could_doit(player, exit, A_LOCK))
    lock=1;

  /* only NPCs who control the destination room may pass through exit */
  if(!lock && RESTRICT_TELEPORT && !Builder(player) && db[exit].link != HOME &&
     Typeof(player) == TYPE_THING && !controls(player, db[exit].link,
     POW_MODIFY))
    lock=1;

  /* In a restricted zone, players can only go through exits that have
     a return exit pointing back to the source room */
  if(!lock && !Builder(player) && is_zone(exit, ZONE_RESTRICTED) &&
     db[exit].link != HOME && !controls(player, exit, POW_MODIFY)) {
    DOLIST(a, db[db[exit].link].exits)
      if(db[a].link == db[player].location)
        break;
    if(a == NOTHING)
      lock=1;
  }

  /* Locked */
  if(lock) {
    ansi_did_it(player, exit, A_FAIL, "You can't go that way.",
                A_OFAIL, NULL, A_AFAIL);
    return;
  }

  show_exitsucc(player, exit);

  enter_room(player, db[exit].link);
  show_exitdrop(player, exit);
}

/* internal: moves an object or exit to a new location */
void enter_room(dbref player, dbref loc)
{
  dbref old=db[player].location, *ptr;
  int invis=0, dead=0, money=1;

  if(loc == HOME)
    loc=db[player].link;

  /* check object validity */
  if(Typeof(player) == TYPE_ROOM || Typeof(player) == TYPE_ZONE) {
    notify(player, tprintf("%ss are not allowed to move.",
           typenames[Typeof(player)]));
    return;
  }

  /* when moving objects to the location they're already in.. */
  if(old == loc) {
    if(loc == NOTHING)
      return;
    /* move object to front of stack */
    if(Typeof(player) == TYPE_EXIT) {
      db[loc].exits = remove_first(db[loc].exits, player);
      PUSH(player, db[loc].exits);
    } else {
      db[loc].contents = remove_first(db[loc].contents, player);
      PUSH(player, db[loc].contents);
    }

    /* look at the room */
    look_room(player, loc, 1);
    return;
  }

  /* check object/location types */

  if(loc != NOTHING && Typeof(loc) == TYPE_EXIT) {
    log_error(tprintf("Attempt to move object %s(#%d) into exit %s(#%d)",
              db[player].name, player, db[loc].name, loc));
    notify(player, "Cannot move to exit.");
    return;
  } if(loc != NOTHING && Typeof(player) == TYPE_PLAYER &&
       Typeof(loc) == TYPE_PLAYER) {
    log_error(tprintf("Attempt to move player %s(#%d) into player %s(#%d)",
              db[player].name, player, db[loc].name, loc));
    notify(player, "Cannot move to player.");
    return;
  } if(loc != NOTHING && Typeof(loc) == TYPE_ZONE) {
    log_error(tprintf("Attempt to move object %s(#%d) into zone %s(#%d)",
              db[player].name, player, db[loc].name, loc));
    notify(player, "Cannot move to zone.");
    return;
  }

  /* update number of steps */
  if(old != loc && loc != NOTHING && Typeof(player) == TYPE_PLAYER &&
     Typeof(loc) == TYPE_ROOM && !IS(loc, TYPE_ROOM, ROOM_XZONE))
    db[player].data->steps++;

  if(Typeof(player) == TYPE_PLAYER && !(db[player].flags & PLAYER_CONNECT))
    invis=1;

  /* move object to new location */

  /* remove object from old contents list */
  if(old != NOTHING) {
    if(Typeof(player) == TYPE_EXIT)
      db[old].exits = remove_first(db[old].exits, player);
    else
      db[old].contents = remove_first(db[old].contents, player);
    ansi_did_it(player, player, A_MOVE, NULL, (invis || dead)?0:A_OMOVE,
                NULL, A_AMOVE);

    /* execute @leave messages on zones */
    if(loc != NOTHING)
      for(ptr=db[mainroom(old)].zone;ptr && *ptr != NOTHING;ptr++)
        if(!inlist(db[mainroom(loc)].zone, *ptr))
          ansi_did_it(player, *ptr, A_LEAVE, NULL, (invis || dead)?0:A_OLEAVE,
                      NULL, A_ALEAVE);
    if(old != loc)
      ansi_did_it(player, old, A_LEAVE, NULL, invis?0:A_OLEAVE,
                  invis?NULL:dead ? "has died.":"has left.", A_ALEAVE);
  }
  
  /* put object into new contents list */
  db[player].location = loc;
  if(loc != NOTHING) {
    if(Typeof(player) == TYPE_EXIT)
      PUSH(player, db[loc].exits);
    else
      PUSH(player, db[loc].contents);

    if(old != NOTHING) {
      for(ptr=db[mainroom(loc)].zone;ptr && *ptr != NOTHING;ptr++)
        if(!inlist(db[mainroom(old)].zone, *ptr))
          ansi_did_it(player,*ptr,A_ENTER,NULL,invis?0:A_OENTER,NULL,A_AENTER);
    }
    if(old != loc)
      ansi_did_it(player, loc, A_ENTER, NULL, invis?0:A_OENTER,
                  (invis || dead)?NULL:"has arrived.", A_AENTER);
  }
  
  /* autolook */
  look_room(player, loc, 1);

  if(old == loc || loc == NOTHING || Typeof(player) != TYPE_PLAYER ||
     IS(loc, TYPE_ROOM, ROOM_XZONE))
    return;

  /* this is a valid room for special player-only features */

  /* penny check */
  if(!invis && money)
    find_money(player);
}

void find_money(dbref player)
{
  int a=1;

  if(rand_num(50))
    return;
  notify(player, tprintf("You find a %s!", MONEY_SINGULAR));
  giveto(player, a);
}

int verb=0;

/* illusionate exit messages. this is fun. :-) */
char *exitmsg(dbref exit, int typ)
{
  static char rslt[525];
  char buf[100];
  char *p, *q=NULL, *r;
  int a, b;

  strncpy(buf, db[exit].name, 99);
  buf[99]=0;
  if((p=strchr(buf, '<'))) {
    *p=0;
    q=p+1;
    if((r=strchr(q, '>')))
      *r=0;
  } else if((p=strchr(buf, '('))) {
    *p=0;
    q=p+1;
    if((r=strchr(q, ')')))
      *r=0;
  }

  b=strlen(buf);
  p=buf+b; /* truncate end spacing and lowercase all letters */
  while(--p >= buf) {
    if(*p == ' ' && !*(p+1))
      *p=0;
    else
      *p=tolower(*p);
  }

  verb=0;  /* goes */
  if(!*buf) {
    strcpy(rslt, typ?"here":"away");
    verb=1; /* walk */
  } else if(Invalid(db[exit].link)) {
    strcpy(rslt, typ?"here":"home");
  } else if(match_word("north northeast east southeast south southwest west "
                       "northwest aft starboard port", buf)) {
    sprintf(rslt, "%s the %s", typ?"from":"to", buf);
    verb=1;
  } else if(match_word("uphill upward upwards", buf)) {
    strcpy(rslt, "up the hill");
    verb=9;
  } else if(match_word("climb up upstairs above", buf)) {
    sprintf(rslt, "%s the room above", typ?"down from":"up to");
    verb=2; /* climb */
  } else if(match_word("ladder ledges vine vines stairs stairway stairwell",
            buf)) {
    if(typ) {
      sprintf(rslt, "off of the %s", buf);
      verb=10;
    } else {
      sprintf(rslt, "onto the %s", buf);
      verb=2;
    }
  } else if(match_word("forward", buf)) {
    strcpy(rslt, buf);
    verb=9;
  } else if(match_word("hillside", buf)) {
    strcpy(rslt, "along the hillside");
    verb=9;
  } else if(match_word("downhill downward downwards", buf)) {
    strcpy(rslt, "down the hill");
    verb=9;
  } else if(match_word("down downhill downstairs downward downwards", buf)) {
    sprintf(rslt, "%s the room below", typ?"up from":"down to");
    verb=2;
  } else if(match_word("out outside leave exit", buf)) {
    strcpy(rslt, typ?"from outside":"back outside");
    verb=1;
  } else if(match_word("in inside enter entrance cabin house cottage", buf)) {
    strcpy(rslt, typ?"from inside":"inside");
    verb=1;
  } else if(match_word("back", buf)) {
    strcpy(rslt, typ?"from ahead":buf);
    verb=9;
  } else if(match_word("ahead", buf)) {
    strcpy(rslt, typ?"from behind":buf);
    verb=9;
  } else if(match_word("tunnel cave cavern tunnels caverns", buf)) {
    strcpy(rslt, "the cave");
    verb=11; /* 3. follow */
  } else if(match_word("hole", buf)) {
    sprintf(rslt, "%s the %s", typ?"from":"into", buf);
    verb=10;
  } else if(match_word("hall hallway steps trail path footpath walkway "
                       "passage branch branches", buf)) {
    sprintf(rslt, "the %s", buf);
    verb=11;
  } else if(match_word("debris rockslide rockslides", buf)) {
    sprintf(rslt, "%s the %s", typ?"from":"beyond", buf);
    verb=1;
  } else if(match_word("jump below", buf)) {
    sprintf(rslt, "down%s", typ?" from above":"");
    verb=12; /* 4. jump */
  } else if(match_word("gate gates thorns door doorway ridges", buf)) {
    sprintf(rslt, "%s the %s", typ?"from":"through", buf);
    verb=9;
  } else if(match_word("bridge road vine", buf)) {
    sprintf(rslt, "%s the %s", typ?"from":"onto", buf);
    verb=9;
  } else if(match_word("cliff cliffs beach shore river fire", buf)) {
    sprintf(rslt, "up by the %s", buf);
    verb=9;
  } else if(match_word("lake", buf)) {
    sprintf(rslt, "across the %s", buf);
    verb=13; /* 5. flow */
  } else if(match_word("crawl crawlspace shaft", buf)) {
    strcpy(rslt, typ?"in":"");
    verb=14; /* 6. crawl */
  } else if(match_word("foyer", buf)) {
    sprintf(rslt, "%s the %s", typ?"inside from":"back to", buf);
    verb=1;
  } else if(strchr(buf, ' ')) {
    sprintf(rslt, "%s %s", typ?"from":"into", buf);
    verb=9;
  } else if((!q || !*q) && !Invalid(db[exit].link)) {
    sprintf(rslt, "%s %s", typ?"from":"into", db[db[exit].link].name);
    verb=1;
  } else {
    sprintf(rslt, "%s the %s", typ?"from":"into", buf);
    verb=9;
  }

  if(q && *q && (verb & 8)) {
    static char *dirtable[]={"N", "north", "NE", "northeast", "E", "east",
                     "SE", "southeast", "S", "south", "SW", "southwest",
                     "W", "west", "NW", "northwest", ""};

    for(a=0;*dirtable[a];a+=2)
      if(!strcasecmp(q, dirtable[a])) {
        sprintf(rslt+strlen(rslt), "%s%s the %s", *rslt?" ":"",
                typ?(verb==11?"from":"in"):"to", dirtable[a+1]);
        break;
      }
  }
  if(verb & 8)
    verb &= ~8;

  strcat(rslt, ".");
  return rslt;
}

void show_exitsucc(dbref player, dbref exit)
{
  char buf[525], buf2[525], temp[200];
  static char *walktable[]={"goes", "goes", "walk", "walks", "climb", "climbs",
                    "follow", "follows", "jump", "jumps", "flow", "flows",
		    "crawl", "crawls"};

  strcpy(temp, exitmsg(exit, 0));
  sprintf(buf, "You %s %s", walktable[verb*2], temp);
  sprintf(buf2, "%s %s", walktable[(verb*2)+1], temp);

  ansi_did_it(player, exit, A_SUCC, buf, A_OSUCC, buf2, A_ASUCC);
}

void show_exitdrop(player, exit)
dbref player;
dbref exit;
{
  char buf[525], temp[200];
  static char *walktable[]={"comes", "walks", "climbs", "follows", "jumps",
                            "flows", "crawls"};
  dbref thing=NOTHING;

  if(db[exit].link != HOME)
    DOLIST(thing, db[db[exit].link].exits)
      if(db[thing].link == db[exit].location)
        break;

  if(thing == NOTHING)
    sprintf(buf, "comes from %s.", db[db[exit].location].name);
  else {
    strcpy(temp, exitmsg(thing, 1));
    sprintf(buf, "%s %s", walktable[verb], temp);
  }

  ansi_did_it(player, exit, A_DROP, NULL, A_ODROP, buf, A_ADROP);
}


void do_teleport(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  dbref thing=player, loc;
  
  /* get victim & destination */
  if((loc=match_thing(player, *arg2?arg2:arg1, MATCH_EVERYTHING|MATCH_HOME)) ==
     NOTHING)
    return;
  if(*arg2 && (thing=match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
    return;

  /* check global restriction */
  if(RESTRICT_TELEPORT && !power(player, POW_TELEPORT) &&
     (!Builder(player) || player != thing)) {
    notify(player, "Restricted command.");
    return;
  }

  /* check object type */
  if(Typeof(thing) == TYPE_ROOM) {
    notify(player, "Rooms are not allowed to move.");
    return;
  } if(Typeof(thing) == TYPE_ZONE) {
    notify(player, "Zones are not allowed to move.");
    return;
  } if(Typeof(thing) == TYPE_EXIT && loc == HOME) {
    notify(player, "Can't reach home.");
    return;
  } if(loc != HOME && ((Typeof(loc) == TYPE_EXIT && db[thing].location !=
       db[loc].location) || Typeof(loc) == TYPE_ZONE)) { 
    notify(player, "That is not a valid destination.");
    return;
  } if(loc != HOME && Typeof(loc) == TYPE_PLAYER && (Typeof(thing) ==
       TYPE_EXIT || Typeof(thing) == TYPE_PLAYER)) {
    notify(player, tprintf(
           "Object type %s may not teleport into object type %s.",
           typenames[Typeof(thing)], typenames[Typeof(loc)]));
    return;
  }

  /* check permissions for victim */
  if(!controls(player, thing, POW_TELEPORT) && !controls(player,
     db[thing].location, POW_TELEPORT)) {
    notify(player, "Permission denied.");
    return;
  }

  /* homeward bound */
  if(loc == HOME) {
    notify(thing, "You awake to find yourself back home...");
    enter_room(thing, loc);
    return;
  }

  /* destination is exit */
  if(Typeof(loc) == TYPE_EXIT) {
    char buf[20];

    sprintf(buf, "#%d", loc);
    do_move(thing, buf);
    return;
  }

  /* check destination */
  if(!controls(player, loc, POW_TELEPORT) && !power(player, POW_TELEPORT) &&
     !IS(loc, TYPE_ROOM, ROOM_JUMP_OK)) {
    notify(player, "Permission denied.");
    return;
  }

  enter_room(thing, loc);
}

/* enter object in room */
void do_enter(dbref player, char *what)
{
  dbref thing;

  /* check permissions */
  if(Typeof(player) == TYPE_EXIT || Typeof(player) == TYPE_ROOM ||
     Typeof(player) == TYPE_ZONE) {
    notify(player, "Only objects may enter things.");
    return;
  }

  if((thing = match_thing(player, what, MATCH_CON)) == NOTHING)
    return;

  if(Typeof(thing) != TYPE_THING) {
    notify(player, "You may only enter things.");
    return;
  }

  if(player == thing || !could_doit(player, thing, A_ELOCK) ||
     (!controls(player, thing, POW_TELEPORT) && !IS(thing, TYPE_THING,
     THING_ENTER_OK))) {
    ansi_did_it(player, thing, A_EFAIL, "You can't enter that.", A_OEFAIL,
                NULL, A_AEFAIL);
    return;
  } if(!could_doit(player, db[player].location, A_LLOCK)) {
    ansi_did_it(player, db[player].location, A_LFAIL,
                "You can't leave the current room.", A_OLFAIL, NULL, A_ALFAIL);
    return;
  }

  enter_room(player, thing);
}          

/* leave an object */
void do_leave(player)
dbref player;
{
  if(Typeof(player) == TYPE_EXIT || Typeof(player) == TYPE_ROOM ||
     Typeof(player) == TYPE_ZONE || Typeof(db[player].location) == TYPE_ROOM ||
     db[db[player].location].location == NOTHING) {
    notify(player, "You can't leave.");
    return;
  }       

  if(!could_doit(player, db[player].location, A_LLOCK) ||
     (Typeof(player) == TYPE_THING && !controls(player,
     db[db[player].location].location, POW_MODIFY)))
    ansi_did_it(player,db[player].location,A_LFAIL,"You can't leave.",
                A_OLFAIL,NULL,A_ALFAIL);
  else
    enter_room(player, db[db[player].location].location);
}

/* move exit/object to front of stack in room */
void do_push(player, arg1)
dbref player;
char *arg1;
{
  dbref thing, loc;

  if((thing=match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
    return;

  loc=db[thing].location;
  if(Invalid(loc) || !controls(player, loc, POW_MODIFY)) {
    notify(player, "Permission denied.");
    return;
  } if(thing == loc || Typeof(thing) == TYPE_ROOM ||
       Typeof(thing) == TYPE_ZONE) {
    notify(player, "You cannot push that object anywhere.");
    return;
  }

  /* push it */
  if(Typeof(thing) == TYPE_EXIT) {
    db[loc].exits = remove_first(db[loc].exits, thing);
    PUSH(thing, db[loc].exits);
  } else {
    db[loc].contents = remove_first(db[loc].contents, thing);
    PUSH(thing, db[loc].contents);
  }

  if(!Quiet(player))
    notify(player, "Pushed forward in stack.");
}

/* remove the first occurence of what in list headed by first */
dbref remove_first(first, what)
dbref first, what;
{
  dbref prev;

  /* special case if it's the first one */
  if(first == what)
    return db[first].next;

  /* have to find it */
  DOLIST(prev, first) {
    if(db[prev].next == what) {
      db[prev].next = db[what].next;
      return first;
    }
  }
  return first;
}

/* moves in the direction of a random exit from current location */
void do_run(dbref player)
{
  dbref exit=random_exit(player, NOTHING);

  if(exit == NOTHING)
    notify(player, "There is nowhere to run.");
  else
    do_move(player, db[exit].name);
}
