/* db/attrib.c */
/* Database attribute maintainence procedures */

#include "externs.h"

ATTR attr[]={
#define DOATTR(num, var, name, flags) {name, NOTHING, num, flags},
#define VREGS
#include "attrib.h"
#undef VREGS
#undef DOATTR
{NULL, 0, 0, 0}};

// Attribute Procedure Testing:
//
// #define DEBUG_ATTR 1


/* lists all attributes to player */
void list_attributes(dbref player, char *name)
{
  int a, b;

  if(*name) {
    b=atoi(name);
    for(a=0;attr[a].num;a++)
      if(attr[a].num == b || !strcasecmp(attr[a].name, name)) {
        notify(player, tprintf("%4d: %-15.15s %s", attr[a].num,
               attr[a].name, unparse_attrflags(attr[a].flags)));
        return;
      }
    notify(player, "Attribute title not found.");
    return;
  }

  notify(player, "Numerical list of attributes and their hardcoded options:");
  notify(player, "");
  for(a=0;attr[a].num;a++)
    notify(player, tprintf("%4d: %-15.15s %s", attr[a].num,
           attr[a].name, unparse_attrflags(attr[a].flags)));
}


/* pointer-returning functions */

/* returns the pointer to a builtin attribute given its name */
ATTR *builtin_atr_str(char *str)
{
  static hash *attrhash=NULL;

  if(!attrhash)
    attrhash=make_hashtab(attr, sizeof(ATTR));
  return lookup_hash(attrhash, str);
}

/* returns the pointer to a defined attribute on <obj> given name */
ATTR *atr_str(dbref obj, char *str)
{
  dbref *ptr;
  ATRDEF *k;
  ATTR *a;

#ifdef DEBUG_ATTR
  if(Invalid(obj)) {
    log_error(tprintf("atr_str: invalid attribute specification: #%d.%s",
              obj, str));
    return builtin_atr_str(str);
  }
#endif

  /* search for defined attribute on <obj> or <obj>'s parents (recursive) */
  for(k=db[obj].atrdefs;k;k=k->next)
    if(!strcasecmp(k->atr->name, str))
      return k->atr;
  for(ptr=db[obj].parents;ptr && *ptr != NOTHING;ptr++)
    if((a=atr_str(*ptr, str)))
      return a;

  return builtin_atr_str(str);
}

ATTR *atr_num(dbref obj, unsigned char num)
{
  static unsigned char attrs[256];
  static int init=0;
  ATRDEF *k;
  int a;

  /* builtin */
  if(obj == NOTHING) {
    if(!init) {
      memset(attrs, 255, 256);  /* set all entries to default byte 255 */
      for(a=0;attr[a].num;a++)
        attrs[attr[a].num]=a;
      init=1;
    }
    return (attrs[num] == 255) ? NULL : &(attr[attrs[num]]);
  }

#ifdef DEBUG_ATTR
  if(Invalid(obj)) {
    log_error(tprintf("atr_num: cannot read attribute list on #%d", obj));
    return NULL;
  }
#endif

  for(k=db[obj].atrdefs;k;k=k->next)
    if(k->atr->num == num)
      return k->atr;
  return NULL;
}


/* routines to handle object attribute lists */

/* adds defined attr <num> on <obj> to <thing> */
void atr_add_obj(thing, obj, num, str)
dbref thing, obj;
unsigned char num;
char *str;
{
  ALIST *ptr, *last=NULL;

#ifdef DEBUG_ATTR
  if(Invalid(thing)) {
    log_error(tprintf("atr_add: cannot add attribute to #%d", thing));
    return;
  }
#endif

  db[thing].mod_time=now;
  str=compress(str);

  for(ptr=db[thing].attrs;ptr;ptr=ptr->next) {
    if(ptr->obj == obj && ptr->num == num) {
      if(!*str) {  /* clear attribute */
        if(last)
          last->next=ptr->next;
        else
          db[thing].attrs=ptr->next;
        free(ptr);
        return;
      }

      /* add or set new value on attribute */
      ptr=realloc(ptr, sizeof(ALIST)+strlen(str)+1);
      if(last)
        last->next=ptr;
      else
        db[thing].attrs=ptr;

      strcpy(ptr->text, str);
      return;
    }
    last=ptr;
  }

  if(!*str)  /* nothing to clear? */
    return;

  /* add attribute */
  ptr=malloc(sizeof(ALIST)+strlen(str)+1);
  ptr->next=NULL;
  ptr->obj=obj;
  ptr->num=num;
  strcpy(ptr->text, str);
  if(last)
    last->next=ptr;
  else
    db[thing].attrs=ptr;
}

/* gets defined attr <num> on <obj> from <thing> */
char *atr_get_obj(thing, obj, num)
dbref thing, obj;
unsigned char num;
{
  dbref *objp;
  ALIST *ptr;
  ATTR *a;
  char *str;

#ifdef DEBUG_ATTR
  if(Invalid(thing)) {
    log_error(tprintf("atr_get: cannot get attribute %d from #%d",num,thing));
    return "";
  }
#endif

  if(!num || (obj != NOTHING && !is_a(thing, obj)))
    return "";
  for(ptr=db[thing].attrs;ptr;ptr=ptr->next)
    if(ptr->obj == obj && ptr->num == num)
      return uncompress(ptr->text);

  if(thing == obj || !(a=atr_num(obj, num)))
    return "";

  /* search for attribute inherited from <obj>'s parents (recursive) */
  if(a->flags & INH)
    for(objp=db[thing].parents;objp && *objp != NOTHING;objp++)
      if(*(str=atr_get_obj(*objp, obj, num)))
        return str;

  return "";
}

/* when destroying <obj> or delparenting <thing> from <obj>, remove all
   attributes that have references to <obj> on its children */
void atr_clean(dbref thing, dbref obj)
{
  ALIST *ptr, *last=NULL, *next;
  dbref *objp;

  for(ptr=db[thing].attrs;ptr;ptr=next) {
    next=ptr->next;
    if(ptr->obj == obj) {
      if(last)
        last->next=next;
      else
        db[thing].attrs=next;
      free(ptr);
    } else
      last=ptr;
  }

  /* now process the children in the same manner */
  for(objp=db[thing].children;objp && *objp != NOTHING;objp++)
    atr_clean(*objp, obj);
}

/* copy attributes from source to dest */
void atr_cpy(dbref dest, dbref source)
{                 
  ALIST *ptr, *new, *last=NULL;
  ATRDEF *k, *newk, *lastk=NULL;
  
  /* copy attribute definitions */
  for(k=db[source].atrdefs;k;k=k->next) {
    newk=malloc(sizeof(ATRDEF));
    newk->next=NULL;
    newk->atr=malloc(sizeof(ATTR));
    newk->atr->obj=source;
    newk->atr->num=k->atr->num;
    newk->atr->flags=k->atr->flags;
    SPTR(newk->atr->name, k->atr->name);
    if(lastk)
      lastk->next=newk;
    else
      db[dest].atrdefs=newk;
    lastk=newk;
  }

  /* copy attribute data */
  for(ptr=db[source].attrs;ptr;ptr=ptr->next) {
    if(ptr->obj != NOTHING && source != ptr->obj && !is_a(dest, ptr->obj))
      continue;
    new=malloc(sizeof(ALIST)+strlen(ptr->text)+1);
    new->next=NULL;
    new->obj=(ptr->obj == source)?dest:ptr->obj;
    new->num=ptr->num;
    strcpy(new->text, ptr->text);
    if(last)
      last->next=new;
    else
      db[dest].attrs=new;
    last=new;
  }
}
