/* db/load.c */
/* contains routines to load most types of mud databases */

#include "externs.h"

struct object *db=NULL;
dbref *uzo=NULL;
dbref db_top=0;
dbref db_size=0;
dbref db_init=0;                    
dbref db_tech=0;

int db_format=0;
int db_vers=0;

int dbref_len=1;
int db_pows=NUM_POWS;
int db_flags=0;
extern int mush_atrdefs;

dbref *zonelist=NULL;


/* allocate memory for a new database object */
dbref new_object()
{
  struct object *new_array;
  dbref i;
  
  /* if there are no destroyed objects to use, increase the size of db array */
  if(!game_online || (i=free_get()) == NOTHING) {
    if(game_online && DB_LIMIT > 1 && db_top >= DB_LIMIT)
      return NOTHING;
    i=db_top++;

    /* when creating first object, allocate entire size of db */
    if(!i)
      db_size=db_init-(db_init % 1000);

    if(!i || db_top > db_size) {
      db_size += 1000;
      if(game_online)
        log_main(tprintf("DB: Increasing Database Array Size to %d Objects",
                 db_size));
      if(!(new_array = realloc(db, db_size * sizeof(struct object)))) {
        if(!game_online) {
          perror("Database Allocation");
          exit(0);
        }
        perror("Database Resize");
        db_size -= 1000;
        db_top--;
        return NOTHING;
      }
      db=new_array;
    }
  }

  /* clear it out */
  memset(db+i, 0, sizeof(struct object));

  /* set default object values */
  SPTR(db[i].name, "");
  db[i].location=NOTHING;
  db[i].contents=NOTHING;
  db[i].exits=NOTHING;
  db[i].link=NOTHING;
  db[i].next=NOTHING;
  db[i].owner=GOD;
  db[i].flags=TYPE_GARBAGE;
  
  return i;
}

/* get dbref# from file */
int getref(FILE *f)
{
  int i;

  if((i=getc(f)) == 0xFF)
    return NOTHING;
  if(dbref_len >= 2) {
    i=(i<<8)|getc(f);
    if(dbref_len >= 3) {
      i=(i<<8)|getc(f);
      if(dbref_len == 4)
        i=(i<<8)|getc(f);
    }
  }

  return (i == db_init)?HOME:i;
}

/* reads string with terminating NULL from file */
char *getstring_noalloc(FILE *f)
{
  static char buf[8192];
  char *s=buf;
  
  while(s-buf < 8000 && (*s=getc(f)))
    s++;
  *s='\0';
  return buf;
}


FILE *db_read_file=NULL;

static struct {
  char *type;	// Name of database format
  int vers;	// Highest version number we know how to load
} dbtypes[]={
  {"Unknown", 0},
  {"TinyMush", 10},
  {"TinyMuse", 3},
  {"TinyMuse Zoned", 15},
  {"TinyMare Compressed RAM", 1},
  {"TinyMush-3", 1},
  {0}};

// When to execute runlevel (for each pass below):
//  0: always
//  1: zonelist != NULL
//  2: db_format != F_MARE
//  3: disable_startups == 0

/* Database validation routines */
static struct {
  char *msg;
  int runlevel;
  void (*func)();
} passes[]={
  {"Installing zone hierarchy",		1, install_zones},
  {"Checking database integrity",	2, check_integrity},
  {"Recycling old objects",		0, destroy_old_objs},
  {"Activating startups",		3, start_game},
  {"Flushing queue",			3, flush_queue},
  {0}};

/* TinyMare-specific configuration rules */
struct db_rule db_rule[]={
  { 1, "configuration setting",	load_config, save_config},
//{ 2, "accounting information",
  {10, "semaphore message",	load_semaphores, save_semaphores},
  {11, "+party group",		load_parties, save_parties},
  {20, "town sector",		load_towns, save_towns},
  {21, "storyline chapter",	load_chapters, save_chapters},
  {22, "levelquest",		load_levelquests, save_levelquests},
//{30, "spell",			load_spells, save_spells},
//{31, "skill",			load_skills, save_skills},
//{32, "event item",		load_event, save_event},
  { 0}};

/* return string containing database type and version */
char *get_db_version()
{
  static char buf[50];

  strcpy(buf, dbtypes[db_format].type);
  if(db_format && db_vers)
    sprintf(buf+strlen(buf), " v%d", db_vers);
  return buf;
}

/* print an error message to the screen and abort */
void db_error(char *message)
{
  char buf[100];

  log_main("--");
  log_main("netmare: Database Load Error...");
  sprintf(buf, "netmare: %.90s", message);
  log_main(buf);
  exit(0);
}

/* create a new database from scratch */
void create_db()
{
  dbref i;

  /* create initial room */
  i=new_object();
  WPTR(db[i].name, "Limbo");
  db[i].flags=TYPE_ROOM;
  db[i].location=i;
  db[i].create_time=now;

  /* create a placeholder for the GOD character */
  recycle(new_object());
}


// Non-blocking load portion of the database file.
// Returns 0 if there is still more to load, 1 when finished.
//
int load_database()
{
  static int state=0;  /* 0=load db header, 1=load objects, 2=do passes */
  static int i, old=0;
  int j;

  /* db loaded, continue with next pass */
  if(state == 2) {
    static int pass=0, index=0;

    /* End of passes? Finalize game. */
    if(!passes[index].msg) {
      if(disable_startups)
        halt_object(AMBIGUOUS);  /* Clear command queue */
      return 1;
    }

    /* check if we are to do this pass */
    if(!passes[index].runlevel ||
       (passes[index].runlevel == 1 && zonelist) ||
       (passes[index].runlevel == 2 && db_format != F_MARE) ||
       (passes[index].runlevel == 3 && !disable_startups)) {

      log_main(tprintf("Pass %2d: %s.", ++pass, passes[index].msg));
      passes[index].func();
    }

    index++;
    return 0;
  }

  /* Load database header */
  if(state == 0) {
    load_db_header(db_read_file);

    log_main(tprintf(" Type: %s", get_db_version()));
    if(db_init)
      log_main(tprintf(" Objs: %d", db_init));
    log_main("");

    /* Tinymush global attribute definitions in header */
    if(mush_atrdefs) {
      log_main(tprintf("Loaded %d attribute definition%s.", mush_atrdefs,
               mush_atrdefs==1?"":"s"));
      log_main("");
    }

    /* Make sure we have a valid db version */
    if(!db_format) {
      log_main("netmare: Unknown database file format. Aborting...");
      log_main("");
      exit(0);
    } if(db_format && db_vers > dbtypes[db_format].vers) {
      log_main("netmare: Cannot load database: Unknown DB version.");
      log_main(tprintf("netmare: I only know about %s versions up to %d.",
               dbtypes[db_format].type, dbtypes[db_format].vers));
      log_main("netmare: Please upgrade to a later release of TinyMARE.");
      log_main("");
      exit(0);
    } if(db_flags & MUSH_GDBM) {
      log_main("netmare: Cannot load database: Not a flat file format.");
      log_main("netmare: Please resave your TinyMUSH database properly.");
      log_main("");
      exit(0);
    }

    /* Select dbref byte-length */
    dbref_len=(db_init < 0xFF)?1:(db_init < 0xFF00)?2:(db_init < 0xFF0000)?3:4;

    /* Initialize zone conversion */
    if(db_format == F_MUSE_ZONED || db_format == F_MUSH3)
      zonelist=calloc(db_init, sizeof(dbref));

    /* Read TinyMare configuration */
    if(db_format == F_MARE) {
      int a, b, count;

      while((a=fgetc(db_read_file)) > 0) {
        for(b=0;db_rule[b].type;b++)
          if(a == db_rule[b].type)
            break;
        if(!db_rule[b].type)
          db_error(tprintf("Unrecognized configuration rule: %d", a));
        count=db_rule[b].load(db_read_file);
        if(count == -1)
          log_main(tprintf("Loaded %s.", db_rule[b].text));
        else if(count)
          log_main(tprintf("Loaded %d %s%s.", count, db_rule[b].text,
                   count==1?"":"s"));
      }
      log_main("");
    }

    /* initialize wtime(), weather, database accounting */
    if(!mare_test_db)
      init_config();

    state=1;
    return 0;
  }

  /* Load DB Objects */

  while(1) { /* load database routine */
    i = db_read_object(i, db_read_file)+1;  /* returns 0 if last object read */
    if((i-1)/1000 > old) {
      while((i-1)/1000 > old) {
        old++;
        if(db_init < 20000 || !(old % 10))
          log_main(tprintf("Reading database object #%d000%s%s%s", old,
                   (old<10)?".":"", (old<100)?".":"", (old<1000)?".":""));
      }
    }
    if(!(i%1000))  /* Preempt loading every 1000 objects */
      break;
  }

  if(i)
    return 0;  /* database not finished loading yet */

  /* finish database object loading and run startup procedures */
  if(Typeof(GOD) != TYPE_PLAYER)
    db_error(tprintf("God account %s(#%d) is not a Player. Aborting...",
             db[GOD].name, GOD));

  fclose(db_read_file);
  log_main("");
  log_main("Database successfully loaded.");
  log_main("");

  /* free temporary space used during conversion */
  if(mush_atrdefs)
    mush_free_atrdefs();

  /* check important configuration dbref's */
  if(Invalid(PLAYER_START) || Typeof(PLAYER_START) != TYPE_ROOM)
    PLAYER_START=0;
  if(Invalid(GAME_DRIVER) || Typeof(GAME_DRIVER) == TYPE_GARBAGE)
    GAME_DRIVER=0;
  if(Invalid(DUNGEON_CELL) || Typeof(DUNGEON_CELL) != TYPE_ROOM)
    DUNGEON_CELL=0;

  /* make sure GOD is still a wizard */
  if(class(GOD) != CLASS_WIZARD)
    initialize_class(GOD, CLASS_WIZARD);

  /* disconnect all previously connected players only on first startup */
  if(nstat[NS_EXEC] == 1)
    for(j=0;j<db_top;j++)
      if(IS(j, TYPE_PLAYER, PLAYER_CONNECT))
        announce_disconnect(j);

  /* move on to db passes */
  state=2;
  return 0;
}

/* Determine database type by reading first few lines */
void load_db_header(FILE *f)
{
  char buf[1024];

  while(1) {
    switch(getc(f)) {
    case EOF:
      return;  /* No valid header found! */
    case '+':
      switch(getc(f)) {
      case 'A':
        mush_get_atrdefs(f);  /* Attribute Definition */
        continue;
      case 'S':
        fgets(buf, 1024, f);
        db_init=atoi(buf);  /* DB Size */
        continue;
      case 'T':
        db_format=F_MUSH3; /* TinyMush v3.0 Header Format */
        fgets(buf, 1024, f);
        db_vers=atoi(buf) & 0xff;
        db_flags=atoi(buf) >> 8;
        continue;
      case 'V':
        db_format=F_MUSH; /* TinyMush v2.0 Header Format */
        fgets(buf, 1024, f);
        db_vers=atoi(buf) & 0xff;
        db_flags=atoi(buf) >> 8;
        continue;
      case 'X':
        db_format=0; /* TinyMux Header Format - unsupported */
        fgets(buf, 1024, f);
        db_vers=atoi(buf) & 0xff;
        db_flags=atoi(buf) >> 8;
        continue;
      default:
        fgets(buf, 1024, f);
        continue;
      }
    case '-':
      switch(getc(f)) {
      case 'R':
        /* record #players online (currently unused) */
      default:
        fgets(buf, 1024, f);
        continue;
      }
    case '@':
      db_format=F_MUSE;  /* TinyMuse Format */
      fgets(buf, 1024, f);
      db_vers=atoi(buf) & 0xff;
      continue;
    case '~':
      fgets(buf, 1024, f);
      db_init=atoi(buf);  /* DB Size */
      continue;
    case '!':
      ungetc('!', f);
      return;
    case '&':
      db_format=F_MUSE_ZONED;  /* Change to TinyMuse Zoned */
      rewind(f);  /* get TinyMuse configuration parameters */
      return;
    case 'M':
      fread(buf, 3, 1, f);
      if(strncmp(buf, "ARE", 3))
        continue;
      db_format=F_MARE;  /* TinyMare Compressed RAM */
      db_vers=fgetc(f);
      db_init=getnum(f, 4);
      db_flags=getnum(f, 4);
      db_pows=fgetc(f);
      return;
    default:
      fgets(buf, 1024, f);
    }
  }
}

int db_read_object(dbref i, FILE *f)
{
  eqtech *mgk, *mpt=NULL;
  eqlist *entry, *last;
  int a, b, c, d, newobj;
  char *ptr, *max;

  /* old database versions */
  switch(db_format) {
  case F_MUSH: case F_MUSH3:
    return mush_db_read_object(i, f);
  case F_MUSE: case F_MUSE_ZONED:
    return muse_db_read_object(i, f);
  }

  /* read in object */
  speaker=newobj=getref(f);
  if(newobj == NOTHING)  /* end of database */
    return NOTHING;

  if(newobj < i || newobj >= db_init || (!i && newobj)) {
    /* something went wrong on last obj# */
    if(i > 0)
      db_error(tprintf("Error loading object: #%d.", i-1));
    else
      db_error("Cannot find first record on database.");
  }

  /* we're skipping objects here, fill in the ones we missed */
  if(newobj > i)
    for(a=i;a<newobj;a++)
      new_object();

  /* read it in */
  new_object();

  getstring(f, db[newobj].name);
  db[newobj].location=getref(f);
  db[newobj].contents=getref(f);
  db[newobj].exits=getref(f);
  db[newobj].link=getref(f);
  db[newobj].next=getref(f);
  db[newobj].owner=getref(f);
  db[newobj].flags=getnum(f, 4);

  if(Typeof(newobj) == TYPE_EXIT)
    db[newobj].plane=-1;  /* exits can be seen in any plane by default */

  db[newobj].pennies=getnum(f, 4);

  if(db_flags & DB_LONGINT) {
    db[newobj].create_time=getlong(f);
    db[newobj].mod_time=getlong(f);
  } else {
    db[newobj].create_time=getnum(f, 4);
    db[newobj].mod_time=getnum(f, 4);
  }

  db[newobj].parents=getlist(f);
  db[newobj].children=getlist(f);
  db[newobj].relay=getlist(f);
  db[newobj].zone=getlist(f);

  /* read in player-specific data structure */
  if(Typeof(newobj) == TYPE_PLAYER) {
    db[newobj].data=malloc(sizeof(struct playerdata)+(db_pows+3)/4);
    db[newobj].data->class=fgetc(f)+1;
    db[newobj].data->rows=fgetc(f);
    db[newobj].data->cols=fgetc(f);

    a=fgetc(f);
    db[newobj].data->tzdst=a >> 6;
    db[newobj].data->tz=a >> 1;
    db[newobj].data->gender=a;

    db[newobj].data->term=getnum(f, 2);
    db[newobj].data->steps=getnum(f, 4);
    db[newobj].data->sessions=getnum(f, 4);
    db[newobj].data->age=getnum(f, 4);

    if(db_flags & DB_LONGINT) {
      db[newobj].data->last=getlong(f);
      db[newobj].data->lastoff=getlong(f);
    } else {
      db[newobj].data->last=getnum(f, 4);
      db[newobj].data->lastoff=getnum(f, 4);
    }

    // Read in password and powers. If the number of powers stored does not
    // match, reset to the defaults for that particular class.

    fread(db[newobj].data->pass, (db_pows+3)/4+10, 1, f);
    if(db_pows != NUM_POWS || db[newobj].data->class >= NUM_CLASSES) {
      db[newobj].data=realloc(db[newobj].data, sizeof(struct playerdata) +
                              (NUM_POWS+3)/4);
      initialize_class(newobj, db[newobj].data->class);
    }
  }

  /* read bitmap */
  if((db[newobj].bitmaplen=getnum(f, 2)) > 0) {
    db[newobj].bitmap=malloc(db[newobj].bitmaplen);
    max=db[newobj].bitmap + db[newobj].bitmaplen;
    for(ptr=db[newobj].bitmap;ptr < max;ptr++)
      *ptr=getc(f);
  }

  /* read techniques */
  mpt=NULL;
  c=fgetc(f);
  for(d=0;d<c;d++) {
    mgk=malloc(sizeof(eqtech));
    mgk->next=NULL;
    mgk->type=fgetc(f);

    last=NULL;
    b=getnum(f, 2);
    for(a=0;a<b;a++) {
      entry=malloc(sizeof(eqlist));
      entry->next=NULL;
      entry->num=getnum(f, 2);
      entry->level=fgetc(f);
      entry->value=getnum(f, 4);
      if(!last)
        mgk->list=entry;
      else
        last->next=entry;
      last=entry;
    }

    if(!mpt)
      db[newobj].tech=mgk;
    else
      mpt->next=mgk;
    mpt=mgk;
  }

  /* read attribute list for item */ 
  get_attrs(f, newobj);
  get_atrdefs(f, newobj);

  /* misc variables */
  if(Typeof(newobj) == TYPE_PLAYER)
    add_player(newobj);
  if(IS(newobj, TYPE_ZONE, ZONE_UNIVERSAL))
    PUSH_L(uzo, newobj);

  return newobj;
}

/* load object lists from database */
dbref *getlist(FILE *f)
{
  dbref a, b, *list=NULL;

  if(!(a=getref(f)+1))
    return NULL;

  list=malloc((a+1)*sizeof(dbref));
  for(b=0;b<a;b++)
    list[b]=getref(f);
  list[a]=NOTHING;
  return list;
}

/* load attribute definitions for object */
void get_atrdefs(FILE *f, dbref thing)
{
  ATRDEF *new, *last=NULL;
  int num=0;
  
  while(1) {
    if(!(num=fgetc(f)))
      return;
    new=malloc(sizeof(ATRDEF));
    new->next=NULL;
    new->atr=malloc(sizeof(ATTR));
    new->atr->obj=thing;
    new->atr->num=num;
    new->atr->flags=getnum(f, 2);
    getstring(f, new->atr->name);

    if(last)
      last->next=new;
    else
      db[thing].atrdefs=new;
    last=new;
  }
}

/* read attribute list */
void get_attrs(FILE *f, dbref thing)
{
  ALIST *new, *last=NULL;
  int num, obj;
  char *s;
  
  while(1) {
    if(!(num=fgetc(f)))
      return;
    obj=getref(f);
    s=getstring_noalloc(f);
    new=malloc(sizeof(ALIST)+strlen(s)+1);
    new->next=NULL;
    new->num=num;
    new->obj=obj;
    strcpy(new->text, s);

    if(last)
      last->next=new;
    else
      db[thing].attrs=new;
    last=new;
  }
}
