/* game/predicates.c */
/* predicates for testing various conditions */

#include "externs.h"


/* returns true (1) if <who> can see <atr> on <what> */
int can_see_atr(who, what, atr)
dbref who, what;
ATTR *atr;
{
  if((atr->flags & AF_HIDDEN) && who != GOD)
    return 0;
  if(!(atr->flags & AF_OSEE) && !controls(who, what, POW_EXAMINE) &&
     !(db[what].flags & VISIBLE))
    return 0;
  if(atr->flags & AF_DARK) {
    if(atr->obj == NOTHING && !power(who, POW_SECURITY))
      return 0;
    if(atr->obj != NOTHING && !(power(who, POW_SECURITY) || ((atr->obj == what)
       && controls(who, what, POW_SECURITY))))
      return 0;
  }
  return 1;
}

/* returns 1 if <who> can Set <atr> on <what> */
int can_set_atr(who, what, atr)
dbref who, what;
ATTR *atr;
{
  if(!controls(who, what, POW_MODIFY))
    return 0;
  if(!can_see_atr(who, what, atr))
    return 0;
  if(atr->flags & AF_WIZ) {
    if(atr->obj == NOTHING && !power(who, POW_WIZATTR))
      return 0;
    if(atr->obj != NOTHING && !controls(who, atr->obj, POW_WIZATTR))
      return 0;
  }
  return 1;
}

/* internal function used to find the main container of a nest of objects */
int mainroom(dbref player)
{
  int dep, var=player;

  if(Invalid(player) || Invalid(db[player].location))
    return player;

  for(dep=0;dep<30;dep++) {
    if(db[var].location == var)
      return var;
    var=db[var].location;
    if(Invalid(var))
      return db[player].location;
    if(Typeof(var) == TYPE_ROOM)
      return var;
  }

  return db[player].location;
}

/* check if player can locate <it> by using his powers */
int locatable(dbref player, dbref it)
{
  dbref room;

  /* No sense if trying to locate a bad object. */
  if(Invalid(it))
    return 0;

  /* Succeed if we can examine the target, if we are the target,
   * or if we can examine the location. */

  if(db[player].location == db[it].location || db[player].location == it ||
     controls(player, it, POW_FUNCTIONS) ||
     controls(player, db[it].location, POW_FUNCTIONS))
    return 1;

  /* Succeed if we control the containing room or if it is a findable
   * player and the containing room is also findable. */

  room=mainroom(it);

  if(controls(player, room, POW_FUNCTIONS) ||
     (Typeof(it) == TYPE_PLAYER && !(db[it].flags & DARK) &&
     (IS(room, TYPE_ROOM, ROOM_JUMP_OK|ROOM_ABODE|ROOM_LINK_OK) ||
      (db[room].flags & VISIBLE))))
    return 1;

  /* Succeed if the target is in a room zoned to the inactor. 
   * This does not work on any universal zones owned by the player. */

  if(inlist(db[room].zone, player))
    return 1;

  return 0;  /* We can't do it. */
}

// Returns 1 if player is in thing's location, player is the container for
// thing, or thing is the container for player.
//
int nearby(dbref player, dbref thing)
{
  /* check valid objects */
  if(Invalid(player) || Invalid(thing))
    return 0;

  return (player == db[thing].location || db[player].location == thing ||
          db[player].location == db[thing].location);
}

// Return a random exit visible and accessible by the player.
// If loc != NOTHING, then ignore exits leading to <loc>.
//
dbref random_exit(dbref player, dbref loc)
{
  dbref obj;
  int index, count=0;

  if(Invalid(player) || Invalid(db[player].location))
    return NOTHING;
  DOLIST(obj, db[db[player].location].exits)
    if(db[obj].link != NOTHING && db[obj].link != loc &&
       (controls(player, obj, POW_EXAMINE) || can_see(player, obj)) &&
       could_doit(player, obj, A_LOCK))
      count++;

  index=rand_num(count);
  count=0;
  DOLIST(obj, db[db[player].location].exits)
    if(db[obj].link != NOTHING && db[obj].link != loc &&
       (controls(player, obj, POW_EXAMINE) || can_see(player, obj)) &&
       could_doit(player, obj, A_LOCK))
      if(count++ == index)
        return obj;

  return NOTHING;
}

/* atr_get's first zone in list with attribute set */
char *atr_zone(dbref player, unsigned char num)
{
  dbref *zone;
  char *s;

  for(zone=db[mainroom(player)].zone;zone && *zone != NOTHING;zone++)
    if(*(s=atr_get(*zone, num)))
      return s;
  return "";
}

/* same as above but with function parsing */
char *parse_atr_zone(player, loc, num)
dbref player, loc;
unsigned char num;
{
  dbref *zone;
  char *s;

  for(zone=db[mainroom(loc)].zone;zone && *zone != NOTHING;zone++)
    if(*(s=atr_get(*zone, num)))
      return parse_function(*zone, player, s);
  return "";
}

/* search all zones on <loc> and return 1 if any flag matches */
int is_zone(dbref loc, int flag)
{
  dbref *zone;

  for(zone=db[mainroom(loc)].zone;zone && *zone != NOTHING;zone++)
    if(db[*zone].flags & flag)
      return 1;
  return 0;
}


// Charge Building Expense
//
// Type=0: Charge amount without notification
//      1: Creating item in database, checks POW_FREE and notifies
//      2: Checks amount available but does not subtract
int charge_money(player, cost, type)
dbref player;
int cost, type;
{
  player=db[player].owner;
  if(cost < 1 || (type && power(player, POW_FREE)))
    return 1;
  if(db[player].pennies < cost) {
    if(type)
      notify(player, tprintf("That costs %d %s.", cost, cost==1?
             MONEY_SINGULAR:MONEY_PLURAL));
    return 0;
  }

  if(type != 2)
    db[player].pennies -= cost;
  return 1;
}

/* modifies player's money value */
void giveto(dbref player, int pennies)
{
  player=db[player].owner;

  db[player].pennies+=pennies;

  if(db[player].pennies < 0)
    db[player].pennies=0;
  else if(db[player].pennies > 999999999)
    db[player].pennies=999999999;
}


// Returns gender number (0 to 3) specific to <player>.
//
int get_gender(dbref player)
{
  char gender;

  if(Typeof(player) == TYPE_PLAYER)
    return db[player].data->gender;  /* player's gender flag */

  /* determine gender by first character of attribute */

  gender=tolower(*atr_get(player, A_SEX));
  if(gender == 'm')
    return MALE;
  if(gender == 'f' || gender == 'w')
    return FEMALE;
  if(gender == 'p')
    return PLURAL;
  return NEUTER;
}

// Returns pronoun for gender substitution.
// <Type> may be one of: SUBJ, OBJN, POSS, or APOSS.
//
char *gender_subst(dbref player, int type)
{
  static char *subst[16]={"he",  "she",  "it",  "they",
			  "him", "her",  "it",  "them",
			  "his", "her",  "its", "their",
			  "his", "hers", "its", "theirs"};

  return subst[type+get_gender(player)];
}


char *secure_string(char *str)
{
  char *s;

  while((s=strchr(str,'(')) || (s=strchr(str,')')) || (s=strchr(str,'[')) ||
        (s=strchr(str,']')) || (s=strchr(str,'{')) || (s=strchr(str,'}')) ||
        (s=strchr(str,'$')) || (s=strchr(str,'%')) || (s=strchr(str,',')) ||
        (s=strchr(str,'^')) || (s=strchr(str,';')) || (s=strchr(str,'\n')) ||
        (s=strchr(str,'\t')) || (s=strchr(str,'\\')))
    *s=' ';

  return str;
}

int ok_attribute_name(char *name)
{
  char *s;

  if(!*name || strlen(name) > 63)
    return 0;

  for(s=name;*s;s++)
    if(!(isalpha(*s) || (*s > ' ' && *s <= '9' && *s != '(' && *s != ')') ||
       *s == '_'))
      return 0;
  return 1;
}

int ok_filename(char *name)
{
  char *s=name;
  int dot=1;

  if(!*name || strlen(name) > 250)
    return 0;

  /* main directory . is always safe to read */
  if(!strcmp(name, "."))
    return 1;

  while(*s) {
    if(*s < 33 || *s > 126 || (dot && *s == '.'))
      return 0;
    dot=(*s++ == '/')?1:0;
  }
  return 1;
}

int ok_name(unsigned char *name)
{
  char buf[52], *s;

  if(!*name || *name == '*' || *name == '#' || *name == '!' || *name == ' ' ||
     *name == '.' || *name == '_')
    return 0;

  for(s=name;*s;s++)
    if(*s == '=' || *s == '&' || *s == '|' || *s == '@' ||
       *s < 32 || (*s > 90 && *s < 95) || *s > 122)
      return 0;

  /* check name length */
  strncpy(buf, name, 51);
  buf[51]='\0';
  if((s=strchr(buf, ';'))) {
    *s='\0';
    while(s > buf && *(s-1) == ' ')
      *--s='\0';
  }
  if(!*buf || strlen(buf) > 50 || match_word("me home here quit", buf))
    return 0;

  return 1;
}

int ok_itemname(char *name)
{
  char *s;

  if(strlen(name) < 1 || strlen(name) > 20)
    return 0;
  if(isdigit(*name))
    return 0;

  for(s=name;*s;s++)
    if(*s != '-' && *s != '\'' && *s != '.' && *s != ' ' && (*s < 'A' ||
       (*s > 'Z' && *s < 'a') || *s > 'z'))
      return 0;

  return 1;
}

int ok_player_name(char *name, int typ)
{
  char *s;

  if((!typ && (isdigit(*name) || strlen(name) < 3)) || strlen(name) > 16)
    return 0;
  if(!ok_name(name) || match_word("new all nuper guest who quit jesus satan hell exit player room zone thing garbage the you shit guard ass asshole anus fuck fucker fuckyou bitch bastard penis cock prick dick wang wiener piss pisser schmuck schlong pecker dong balls nuts testes vagina cunt pussy twat poontang clit clitoris clint hymen breast breasts tit tits knocker knockers bonker bonkers honker honkers boob boobs hooter hooters", name))
    return 0;
  if(!strcasecmp(name, MUD_NAME))
    return 0;

  for(s=name;*s;s++)
    if((*s < 48 || (*s > 57 && *s < 65)) && *s != '.' && *s != '-')
      return 0;

  if(string_prefix(name, GUEST_PREFIX))
    return 0;
  if(lookup_player(name) != NOTHING)
    return 0;

  return 1;
}

int ok_password(char *password)
{
  char *s;
  
  if(strlen(password) < 3)
    return 0;

  for(s=password;*s;s++)
    if(*s < 33 || *s > 126)
      return 0;
  return 1;
}                  

char *possname(char *name)
{
  if(!name || !*name)
    return "";
  if(name[strlen(name)-1] == 's')
    return "'";
  return "'s";
}


/* Parse string */
char *parse_function(privs, player, str)
dbref privs, player;
char *str;
{
  static char buf[8192];
  char text[strlen(str)+1];

  if(*str) {
    strcpy(text, str);
    pronoun_substitute(buf, privs, player, text);
  } else
    *buf='\0';
  return buf;
}

/* Parse string into specified buffer space (8192 bytes) */
void pronoun_substitute(buf, privs, player, str)
dbref privs, player;
char *buf, *str;
{
  char temp[8192], c, *p, *r, *s;
  int var, pos=0, color=default_color;

#ifdef DEBUG
  /* catch invalid calls to the pronoun parser */
  if(Invalid(privs) || Invalid(player)) {
    log_error(tprintf("FUNC: #%d<-#%d %s", privs, player, str));
    *buf='\0';
    return;
  }
#endif

  for(s=str;*s && pos < 7999;s++) {
    /* Function Substitution */
    if(*s == '[') {
      r=s++; p=temp;
      while(*s && *s != '(')
        *p++=*s++;
      *p='\0';
      if(*s++ == '(' && function(&s, temp, privs, player)) {
        temp[8000-pos]=0;
        strcpy(buf+pos, temp);
        pos=strlen(buf);
        if(*s != ']')
          s--;
        continue;
      }
      s=r;
    }

    /* Environment Variable Substitution */
    if(*s == '%') {
      c=*(++s); *temp=0;
      switch(tolower(c)) {
      case '0' ... '9':
        strcpy(temp, wptr[c-'0']);

        /* secure string */
        for(var=0;temp[var];var++)
          if(temp[var] == ';' || temp[var] == ',')
            temp[var]=' ';
        break;
      case 'a':
        strcpy(temp, gender_subst(player, APOSS)); break;
      case 'b':
        buf[pos++]=' '; break;
      case 'g':
        buf[pos++]='\a'; break;
      case 'l':
        sprintf(temp, "#%d", db[player].location); break;
      case 'n':
        strcpy(temp, db[player].name); break;
      case 'o':
        strcpy(temp, gender_subst(player, OBJN)); break;
      case 'p':
        strcpy(temp, gender_subst(player, POSS)); break;
      case 'r':
        buf[pos++]='\n'; break;
      case 's':
        strcpy(temp, gender_subst(player, SUBJ)); break;
      case 't':
        buf[pos++]='\t'; break;
      case 'u':
        strcpy(temp, db[privs].name); break;
      case 'v':
        if(!(var=tolower(*(++s)))) {
          s--;
          break;
        } if(var < 'a' || var > 'z')
          buf[pos++]=*s;
        else
          strcpy(temp, atr_get(privs, 100+var-'a'));

        /* secure string */
        for(var=0;temp[var];var++)
          if(temp[var] == ';' || temp[var] == ',')
            temp[var]=' ';
        break;
      case 'x':
        if(!*++s) {
          s--;
          break;
        }
        var=0;
        p=s;
        while(*p && !isdigit(*p)) {
          temp[var++]=*p;
          if(*(p+1) != '%' || (*(p+2) != 'x' && *(p+2) != 'X'))
            break;
          p+=3;
        }
        if(!*(s=p)) {
          s--;
          *temp='\0';
        } else {
          temp[var]='\0';
          var=color_code(default_color, temp);
          strcpy(temp, textcolor(default_color, var));
          default_color=var;
        }
        break;
      case '!':
        sprintf(temp, "#%d", privs); break;
      case '#':
        sprintf(temp, "#%d", player); break;
      case '$':
        *temp='\0'; break;  /* Reserved for program accumulator */
      case '?':
        strcpy(temp, func_printlev()); break;
      case '|':
        for(p=s+1;*p && *p != '|';p++);
        if(!*p)
          break;
        *p='\0';
        var=color_code(7, s+1);
        strcpy(temp, textcolor(default_color, var));
        default_color=var;
        s=p;
        break;
      default:
        buf[pos++]=*s;
      }

      if(*temp) {
        if(c >= 'A' && c <= 'Z')  /* capitalize first character */
          *temp=toupper(*temp);
        temp[8000-pos]=0;

        strcpy(buf+pos, temp);
        pos=strlen(buf);
      }
      continue;
    }

    if(*s == 27) {  /* escape character */
      buf[pos++]=27;
      if(*(s+1))
        buf[pos++]=(*(++s));
      continue;
    } if(*s == '\\') {  /* literal escape */
      if(!(var=(*(++s)))) {
        s--;
        break;
      }
      buf[pos++]=var;
      continue;
    }

    buf[pos++]=*s;
  }

  buf[pos]='\0';

  /* set color back to default */
  if(color != default_color) {
    strcpy(temp, textcolor(default_color, color));
    if(strlen(buf)+strlen(temp) > 7999)
      strcpy(buf+7999-strlen(temp), temp);
    else
      strcat(buf, temp);
    default_color=color;
  }
}


void did_it(player, thing, Succ, emit, Osucc, oemit, Asucc)
dbref player, thing;
unsigned char Succ, Osucc, Asucc;
char *emit, *oemit;
{
  char *s;

  /* Execute Action */
  if(Asucc && *(s=atr_get(thing, Asucc)))
    parse_que(thing, s, player);

  /* Notify Player */
  if(Succ && *(s=atr_parse(thing, player, Succ)))
    notify(player, s);
  else if(emit)
    notify(player, emit);

  /* check lock on @oemit */
  if(!pass_slock(player, 0))
    return;

  /* Notify Others */
  if(Osucc && *(s=atr_parse(thing, player, Osucc)))
    notify_except(db[player].location, player, tprintf("%s %s",
                  db[player].name, s));
  else if(oemit)
    notify_except(db[player].location, player, tprintf("%s %s",
                  db[player].name, oemit));
}

void ansi_did_it(player, thing, atr1, msg1, atr2, msg2, atr3)
dbref player, thing;
unsigned char atr1, atr2, atr3;
char *msg1, *msg2;
{
  char buf[24], *s;

  /* execute actions */
  did_it(player, thing, 0, NULL, atr2, msg2, atr3);

  if(!atr1 && !msg1)
    return;

  /* build output */
  default_color=unparse_color(player, thing, 4);
  if(!atr1 || !*(s=atr_parse(thing, player, atr1)))
    s=msg1;

  if(s && *s) {
    strcpy(buf, textcolor(7, default_color));
    notify(player, tprintf("%s%s%s", buf, s, textcolor(default_color, 7)));
  }
  default_color=7;
}

/* instantly triggers attribute on all objects in the database */
void trigger_all(unsigned char attr)
{
  dbref thing;
  char *s;
  int var;

  // Special precaution "create_time < now" for immediate commands
  // that create new objects in the database.

  for(thing=0;thing<db_top;thing++)
    if(!(db[thing].flags & (GOING|HAVEN)) && (Typeof(thing) !=
       TYPE_PLAYER || (db[thing].flags & PLAYER_CONNECT)) &&
       db[thing].create_time < now && *(s=atr_get(thing, attr))) {
      immediate_que(thing,s,thing);
      for(var=0;var<10;var++)
        *wptr[var]='\0';
    }
}

/* returns 1 if a player is in <thing> or a container in <thing> */
int hold_player(dbref thing)
{
  dbref inside;

  DOLIST(thing, db[thing].contents) {
    if(Typeof(thing) == TYPE_PLAYER)
      return 1;
    DOLIST(inside, thing)
      if(Typeof(inside) == TYPE_PLAYER)
        return 1;
  }
  return 0;
}

/* count total memory used by an object */
int mem_usage(dbref thing)
{
  eqdef *equip; eqtech *eqp; eqlist *mgk;
  ATRDEF *j; ALIST *m;
  int k, L=sizeof(char *), *ptr;

  k = sizeof(struct object)+L;
  if(db[thing].name)
    k += strlen(db[thing].name)+L+1;

  if((ptr=db[thing].parents)) {
    for(;ptr && *ptr != NOTHING;ptr++) k += sizeof(dbref);
    k += sizeof(dbref)+L;
  } if((ptr=db[thing].children)) {
    for(;ptr && *ptr != NOTHING;ptr++) k += sizeof(dbref);
    k += sizeof(dbref)+L;
  } if((ptr=db[thing].relay)) {
    for(;ptr && *ptr != NOTHING;ptr++) k += sizeof(dbref);
    k += sizeof(dbref)+L;
  } if((ptr=db[thing].zone)) {
    for(;ptr && *ptr != NOTHING;ptr++) k += sizeof(dbref);
    k += sizeof(dbref)+L;
  }

  if(db[thing].data)
    k += sizeof(struct playerdata)+(NUM_POWS+3)/4+L;
  if(db[thing].bitmap)
    k += db[thing].bitmaplen+L;

  for(eqp=db[thing].tech;eqp;eqp=eqp->next) {
    k += sizeof(eqtech)+L;
    for(mgk=eqp->list;mgk;mgk=mgk->next)
      k += sizeof(eqlist)+L;
  }

  for(equip=db[thing].equip;equip;equip=equip->next)
    k += sizeof(eqdef)+L;

  /* add in attribute memory */
  for(m=db[thing].attrs;m;m=m->next) {
    k += sizeof(ALIST)+L;
    k += strlen(m->text)+L+1;
  }
  for(j=db[thing].atrdefs;j;j=j->next) {
    k += sizeof(ATRDEF)+L;
    k += strlen(j->atr->name)+L+1;
  }
  return k;
}

/* counts total memory used by player */
int playmem_usage(dbref player)
{
  int a, b=0;

  for(a=0;a<db_top;a++)
    if(db[a].owner == player)
      b += mem_usage(a);
  return b;
}
