/* io/ident.c */
/* identify a user */

#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <sys/errno.h>
#include "externs.h"

// Query the remote host's identity.
// Returns result in 'rslt' with length 32.

void get_ident(char *rslt, int sock)
{
  struct sockaddr_in localaddr, remoteaddr, sin_addr;
  struct timeval timeout;
  fd_set filedes;
  char buf[128], *str, *s;
  int err, fd, len;

  *rslt='\0';	/* start off with an empty result buffer */

  /* set up local/remote addresses */
  len = sizeof(remoteaddr);
  if(getpeername(sock, (struct sockaddr *)&remoteaddr, &len))
    return;
  len = sizeof(localaddr);
  if(getsockname(sock, (struct sockaddr *)&localaddr, &len))
    return;

  /* open socket connection to remote address */
  if((fd=socket(AF_INET, SOCK_STREAM, 0)) == -1)
    return;
  fcntl(fd, F_SETFL, O_NDELAY);
  sin_addr.sin_family = AF_INET;
  sin_addr.sin_addr = localaddr.sin_addr;
  sin_addr.sin_port = 0;
  if(bind(fd, (struct sockaddr *)&sin_addr, sizeof(sin_addr))) {
    close(fd);
    return;
  }

  /* attempt to do socket connection in progress, timeout 2 seconds */
  sin_addr.sin_family = AF_INET;
  sin_addr.sin_addr = remoteaddr.sin_addr;
  sin_addr.sin_port = htons(113); /* identd port */
  connect(fd, (struct sockaddr *)&sin_addr, sizeof(sin_addr));

  FD_ZERO(&filedes);
  FD_SET(fd, &filedes);
  timeout.tv_sec=2;
  timeout.tv_usec=0;
  if(select(fd+1, 0, &filedes, 0, &timeout) < 1) {
    close(fd);
    return;
  }
  len=sizeof(err);
  getsockopt(fd, SOL_SOCKET, SO_ERROR, (char *)&err, &len);
  log_resolv(tprintf("Ident timer: %.2fs%s", 2.0-((double)timeout.tv_sec +
             (double)timeout.tv_usec/1000000), err?" (Refused)":""));
  if(err) {
    close(fd);
    return;
  }

  /* send identification request */
  sprintf(buf, "%d,%d\r\n", ntohs(remoteaddr.sin_port),
          ntohs(localaddr.sin_port));
  write(fd, buf, strlen(buf));

  /* wait for return input */
  timeout.tv_sec=2;
  timeout.tv_usec=0;
  if(select(fd+1, &filedes, 0, 0, &timeout) < 1) {
    close(fd);
    return;
  }
  log_resolv(tprintf("Response timer: %.2fs", 2.0-((double)timeout.tv_sec +
             (double)timeout.tv_usec/1000000)));

  /* read input */
  len=read(fd, buf, 127);
  close(fd);
  if(len <= 2)  /* read error */
    return;

  /* parse input string and search for userid */
  buf[--len]='\0';
  if(buf[--len] == '\r')
    buf[len]='\0';
  s=buf;
  for(len=0;len<4;len++)
    str=parse_up(&s,':');

  if(!str)  /* wanted string was not found */
    return;

  if(*str == ' ')
    str++;
  strncpy(rslt, str, 31);
  rslt[31]='\0';
}
