/* prog/boolexp.c */
/* utilities for parsing TinyMare @lock mechanism */

#include "externs.h"

#define skip_whitespace(x) while(*(x) && isspace(*(x))) (x)++

static int boolexp();
static int boolexp_OPER();
static int boolexp_REF();

dbref parse_player;
dbref parse_object;

int lock_recursion;	// Protection from recursion

// @Lock Accepts the following keys:
//
// Key Type    Form in @Lock Command
// ----------  ------------------------------
// Normal      <object>
// Is          =<object>
// Carry       +<object>
// Ownership   $<object>
// Indirect    @<object>
//             @<object>/<attribute>
// Attribute   <attribute>:<wildcard-pattern>
//             =<attribute>:<wildcard-pattern>
//             +<attribute>:<wildcard-pattern>
//             <object>/<attribute>:<wildcard-pattern>
// Evaluation  [function()]
// Compound    <key> | <key>
//             <key> & <key>
//             !<key>
//             ( <key> )


/* Grabs the next literal word from the string.
   Returns 1 if the word is a function call. */
static int get_word(char *d, char **s, int type)
{
  char *start=d;
  int func=0, dep;

  skip_whitespace(*s);

  while(**s && **s != '&' && **s != '|' && **s != ')' &&
        (!type || (**s != ':' && **s != '/'))) {
    /* Copy text literally between [ ] in functions */
    if(**s == '[') {
      func=1;
      dep=0;
      while(**s) {
        if(**s == '[')
          dep++;
        else if(**s == ']')
          if(!--dep)
            break;
        *d++=*(*s)++;
      }
    }
    *d++=*(*s)++;
  }

  while(d > start && isspace(*(d-1)))	/* eat following whitespace */
    *d--='\0';
  *d='\0';
  return func;
}

/* process an inputted lock string for placement into an attribute */
char *process_lock(dbref player, dbref object, char *arg)
{
  ATTR *attr;
  static char buff[8192];
  char temp[1024], type, *r=buff, *s=arg;
  int func, dep=0;
  dbref thing;

  while(*s) {
    /* copy prefix delimiters */
    while(*s && (*s == '!' || *s == '(' || isspace(*s))) {
      if(!isspace(*s))
        *r++=*s;
      if(*s++ == '(')
        ++dep;
    }

    if(*s == '$' || *s == '+' || *s == '=' || *s == '@')
      type=*r++=*s++;
    else
      type=0;

    /* get the next literal */
    func=get_word(r, &s, 1);

    switch(*s) {
    case '/':  /* an indirect attribute */
      if(type && type != '@') {
        notify(player, tprintf("Invalid key for '%c'-type lock.", type));
        return NULL;
      }
      if((thing=match_thing(player, r, MATCH_EVERYTHING)) == NOTHING)
        return NULL;
      sprintf(r, "#%d", thing);
      r=strchr(r, 0);
      *r++=*s++;

      /* get the attribute portion */
      func=get_word(r, &s, 1);

      if(*s == '/' || (type == '@' && *s == ':')) {
        notify(player, tprintf("Invalid key for '%c'-type lock.", type));
        return NULL;
      } if(!func) {
        if(!(attr=atr_str(thing, r))) {
          notify(player, tprintf("No such attribute '%s' on %s.", r,
                 unparse_object(player, thing)));
          return NULL;
        } if(!can_see_atr(player, thing, attr)) {
          notify(player, tprintf("Cannot see attribute '%s' on %s.",attr->name,
                 unparse_object(player, thing)));
          return NULL;
        } if(type == '@' && !(attr->flags & AF_LOCK)) {
          notify(player, tprintf("Attribute '%s' is not a lock.", attr->name));
          return NULL;
        } if(!can_see_atr(object, thing, attr)) {
          /* Wizard setting a non-wiz lock */
          strcpy(temp, unparse_object(player, object));
          notify(player,tprintf("Warning: %s cannot see attribute '%s' on %s.",
                 temp, attr->name, unparse_object(player, thing)));
        }
      }
      r=strchr(r, 0);
      *r++=*s++;
      get_word(r, &s, 0);
      break;

    case ':':  /* a built-in attribute */
      if(type == '$' || type == '@') {
        notify(player, tprintf("Invalid key for '%c'-type lock.", type));
        return NULL;
      }
      if(!func && !(attr=builtin_atr_str(r)) && type != '@')
        notify(player, tprintf("Warning: No such built-in attribute '%s'.",r));
      r=strchr(r, 0);
      *r++=*s++;
      get_word(r, &s, 0);
      break;

    default:
      if(func)
        break;
      if((thing=match_thing(player, r, MATCH_EVERYTHING)) == NOTHING)
        return NULL;
      sprintf(r, "#%d", thing);

      /* check permissions on Indirect locks */
      if(type == '@') {
        if(!(attr=atr_num(NOTHING, A_LOCK)) || !can_see_atr(player, thing,
           attr)) {
          notify(player, tprintf("Cannot see default lock on %s.",
                 unparse_object(player, thing)));
          return NULL;
        } if(!can_see_atr(object, thing, attr)) {
          /* Wizard setting a non-wiz lock */
          strcpy(temp, unparse_object(player, object));
          notify(player,tprintf("Warning: %s cannot see default lock on %s.",
                 temp, unparse_object(player, thing)));
        }
      }
    }

    r=strchr(r, 0);
    while(*s && *s != '&' && *s != '|') {
      if((*s != ')' && !isspace(*s)) || (*s == ')' && --dep < 0)) {
        notify(player, "I don't understand that key.");
        return NULL;
      }
      if(*s == ')')
        *r++=*s;
      s++;
    }
    if(*s)
      *r++=*s++;
  }

  /* mismatched ()'s */
  if(dep) {
    notify(player, "I don't understand that key.");
    return NULL;
  }

  *r='\0';
  return buff;
}


/* Process a locked attribute: Entry point */
int eval_boolexp(dbref player, dbref object, char *key)
{
  lock_recursion=0;
  parse_player=player;
  parse_object=object;
  return boolexp(object, key);
}

/* recursive internal expression evaluation */
static int boolexp(dbref object, char *key)
{
  char buf[8192], *s=buf;

  if(!*key)
    return 1;
  if(++lock_recursion == 10) {
    log_main(tprintf("Boolean Expr.: Recursion detected in %s(#%d) lock.",
             db[parse_object].name, parse_object));
    notify(db[parse_object].owner, tprintf(
           "Warning: Recursion detected in %s lock.",
           unparse_object(db[parse_object].owner, parse_object)));
  }
  if(lock_recursion >= 10)
    return 0;

  pronoun_substitute(buf, object, parse_player, key);
  return boolexp_OPER(&s);
}

/* perform boolean operations */
static int boolexp_OPER(char **buf)
{
  int lhs, true=1;

  /* unary negation */
  while(**buf && (isspace(**buf) || **buf == '!')) {
    if(**buf == '!')
      true^=1;
    (*buf)++;
  }

  /* nested parentheses */
  if(**buf == '(') {
    (*buf)++;
    lhs=boolexp_OPER(buf) ? true : !true;
    skip_whitespace(*buf);
    if(**buf == ')')
      (*buf)++;
  } else
    lhs=boolexp_REF(buf) ? true : !true;

  skip_whitespace(*buf);

  if(**buf == '&') {
    (*buf)++;
    return (boolexp_OPER(buf) && lhs);
  }

  if(**buf == '|') {
    (*buf)++;
    return (boolexp_OPER(buf) || lhs);
  }

  return lhs;
}

/* determine reference value of lock */
static int boolexp_REF(char **buf)
{
  ATTR *attr;
  char type=0, str[8192], *s=str, *t;
  dbref thing, player;

  get_word(s, buf, 0);  /* retrieve entire key */

  /* Numerical function result */
  if(isdigit(*str) || *str == '-')
    return atoi(str);

  /* Check Lock Type */

  /* Indirect-Type Lock '@' */
  if(*str == '@') {
    if((s=strchr(str, '/')))
      *s++='\0';

    /* get dbref number */
    for(t=str+1;*t && isspace(*t);t++);
    if((thing=match_thing(parse_object, t, MATCH_EVERYTHING|MATCH_QUIET)) ==
       NOTHING)
      return 0;

    /* attribute specified */
    if(s)
      attr=atr_str(thing, s);
    else
      attr=atr_num(NOTHING, A_LOCK);

    if(!attr || !can_see_atr(parse_object, thing, attr) ||
       !(attr->flags & AF_LOCK))
      return 0;
    return boolexp(parse_object, atr_get_obj(thing, attr->obj, attr->num));
  }

  /* Owner-Type Lock '$' */
  if(*str == '$') {
    for(s=str+1;*s && isspace(*s);s++);
    thing=match_thing(parse_object, s, MATCH_EVERYTHING|MATCH_QUIET);
    return (thing != NOTHING && db[thing].owner == db[parse_player].owner);
  }

  /* Object/Attribute:Pattern Match */
  if((s=strchr(str, '/'))) {
    *s++='\0';

    if((thing=match_thing(parse_object, str, MATCH_EVERYTHING|MATCH_QUIET)) ==
       NOTHING)
    return 0;

    if((t=strchr(s, ':')))
      *t++='\0';
    else
      t="?*";  /* match a non-empty attribute */

    if(!(attr=atr_str(thing, s)) || !can_see_atr(parse_object, thing, attr))
      return 0;
    return wild_match(atr_get_obj(thing, attr->obj, attr->num), t, 1, 0);
  }

  /* Determine Lock Type: Const, Carry '+', Is '=' */

  s=str;
  if(*s == '+' || *s == '=') {
    type=*s++;
    skip_whitespace(s);
  }

  if((t=strchr(s, ':'))) {	/* Key = Attribute:Pattern */
    *t++='\0';
    thing=NOTHING;
  } else {			/* Key = Object */
    if((thing=match_thing(parse_object, s, MATCH_EVERYTHING|MATCH_QUIET)) ==
       NOTHING)
    return 0;
  }

  /* Check Player: Const-Type & Is-Type '=' */
  if(type != '+') {
    if(t && (attr=atr_str(parse_player, s)) && can_see_atr(parse_object,
       parse_player, attr)) {
      if(wild_match(atr_get_obj(parse_player, attr->obj, attr->num), t, 1, 0))
        return 1;
    } else if(thing != NOTHING && parse_player == thing)
      return 1;
  }

  /* Check Contents: Const-Type & Carry-Type '+' */
  if(type != '=')
    DOLIST(player, db[parse_player].contents) {
      if(t && (attr=atr_str(player, s)) && can_see_atr(parse_object,
         player, attr)) {
        if(wild_match(atr_get_obj(player, attr->obj, attr->num), t, 1, 0))
          return 1;
      } else if(thing != NOTHING && player == thing)
        return 1;
    }

  return 0;
}
